import os
import json
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.exc import OperationalError
from sqlalchemy.sql import text
from hape.config import Config
from hape.logging import Logging

class HapeConfig:
    logger = Logging.get_logger('hape.config_db')
    _env_loaded = False
    _db_session = None
    _required_env_variables = []

    Config.set_required_env_variable("HAPE_GITLAB_TOKEN")
    Config.set_required_env_variable("HAPE_GITLAB_DOMAIN")
    Config.set_required_env_variable("HAPE_MARIADB_HOST")
    Config.set_required_env_variable("HAPE_MARIADB_USERNAME")
    Config.set_required_env_variable("HAPE_MARIADB_PASSWORD")
    Config.set_required_env_variable("HAPE_MARIADB_DATABASE")

    @staticmethod
    def check_variables():
        Config.logger.debug(f"check_variables()")
        for variable in Config._required_env_variables:
            Config._get_env_value(variable)

    @staticmethod
    def get_db_url():
        Config.logger.debug(f"get_db_url()")
        return f"mysql+pymysql://{HapeConfig.get_mariadb_username()}:{HapeConfig.get_mariadb_password()}@{HapeConfig.get_mariadb_host()}/{HapeConfig.get_mariadb_database()}"

    @staticmethod
    def get_db_session() -> sessionmaker:
        Config.logger.debug(f"get_db_url()")
        if not HapeConfig._db_session:
            try:
                DATABASE_URL = HapeConfig.get_db_url()
                engine = create_engine(DATABASE_URL, echo=True)
                with engine.connect() as connection:
                    connection.execute(text("SELECT 1"))

                HapeConfig._db_session = sessionmaker(autocommit=False, autoflush=False, bind=engine)
                Config.logger.info("Database seassion created successfully.")
            except OperationalError:
                Config.logger.error("Error: Unable to connect to the database. Please check the configuration.")
                raise

        return HapeConfig._db_session()

    @staticmethod
    def get_gitlab_token():
        Config.logger.debug(f"get_gitlab_token()")
        return Config._get_env_value("HAPE_GITLAB_TOKEN")
    
    @staticmethod
    def get_gitlab_domain():
        Config.logger.debug(f"get_gitlab_domain()")
        return Config._get_env_value("HAPE_GITLAB_DOMAIN")
    
    @staticmethod
    def get_mariadb_host():
        Config.logger.debug(f"get_mariadb_host()")
        return Config._get_env_value("HAPE_MARIADB_HOST")

    @staticmethod
    def get_mariadb_username():
        Config.logger.debug(f"get_mariadb_username()")
        return Config._get_env_value("HAPE_MARIADB_USERNAME")

    @staticmethod
    def get_mariadb_password():
        Config.logger.debug(f"get_mariadb_password()")
        return Config._get_env_value("HAPE_MARIADB_PASSWORD")

    @staticmethod
    def get_mariadb_database():
        Config.logger.debug(f"get_mariadb_database()")
        return Config._get_env_value("HAPE_MARIADB_DATABASE")



from alembic import op
import sqlalchemy as sa

# Revision identifiers, used by Alembic.
revision = '001_create_deployment_cost'
down_revision = None
branch_labels = None
depends_on = None

def upgrade():
    op.create_table(
        'deployment_costs',
        sa.Column('id', sa.Integer, primary_key=True, autoincrement=True),
        sa.Column('created_at', sa.BigInteger, nullable=False),
        sa.Column('service_name', sa.String(255), nullable=False),
        sa.Column('pod_cpu', sa.String(50), nullable=False),
        sa.Column('pod_ram', sa.String(50), nullable=False),
        sa.Column('autoscaling', sa.Boolean, nullable=False),
        sa.Column('min_replicas', sa.Integer, nullable=True),
        sa.Column('max_replicas', sa.Integer, nullable=True),
        sa.Column('current_replicas', sa.Integer, nullable=False),
        sa.Column('pod_cost', sa.Float, nullable=False),
        sa.Column('number_of_pods', sa.Integer, nullable=False),
        sa.Column('total_cost', sa.Float, nullable=False),
        sa.Column('cost_unit', sa.String(50), nullable=False),
    )

def downgrade():
    op.drop_table('deployment_costs')



import argparse
from importlib.metadata import version

from hape.argument_parsers.playground_argument_parser import PlaygroundArgumentParser
from hape.argument_parsers.config_argument_parser import ConfigArgumentParser
from hape.argument_parsers.git_argument_parser import GitArgumentParser
from hape.argument_parsers.deployment_cost_argument_parser import DeploymentCostArgumentParser

class MainArgumentParser:

    def create_parser(self):
        parser = argparse.ArgumentParser(
            description="hapeShortLong used to streamline development operations"
        )
        try:
            parser.add_argument("-v", "--version", action="version", version=version("hape"))
        except:
            parser.add_argument("-v", "--version", action="version", version="0.0.0")
        
        subparsers = parser.add_subparsers(dest="command")
        
        PlaygroundArgumentParser().create_subparser(subparsers)
        ConfigArgumentParser().create_subparser(subparsers)
        GitArgumentParser().create_subparser(subparsers)
        DeploymentCostArgumentParser().create_subparser(subparsers)

        return parser
    
    def run_action(self, args):
        
        if args.command == "play":
            PlaygroundArgumentParser().run_action(args)
        elif args.command == "config":
            ConfigArgumentParser().run_action(args)
        elif args.command == "git":
            GitArgumentParser().run_action(args)
        elif args.command == "deployment-cost":
            DeploymentCostArgumentParser().run_action(args)
        else:
            print(f"Error: Invalid command {args.command}")
            exit(1)



import argparse
from importlib.metadata import version

from hape.hape_cli.argument_parsers.init_argument_parser import InitArgumentParser

class MainArgumentParser:

    def create_parser(self):
        parser = argparse.ArgumentParser(description="HAPE Framework CLI")

        try:
            parser.add_argument("-v", "--version", action="version", version=version("hape"))
        except:
            parser.add_argument("-v", "--version", action="version", version="0.0.0")
        
        subparsers = parser.add_subparsers(dest="command")
        
        InitArgumentParser().create_subparser(subparsers)

        return parser

    def run_action(self, args):

        if args.command == "init":
            InitArgumentParser().run_action(args)
        else:
            print(f"Error: Invalid command {args.command}")
            exit(1)



from hape.logging import Logging

import kubernetes.config
import kubernetes.client

class KubernetesService:
    
    _context_dev = ""
    _context_test = ""
    _context_prod = ""

    def __init__(self, context):
        if context == 'dev':
            cluster_context = self._context_dev
        elif context == 'test':
            cluster_context = self._context_test
        elif context == 'prod':
            cluster_context = self._context_prod
        else:
            raise ValueError("Error: Kubernetes context not defined")
        
        kubernetes.config.load_kube_config(context=cluster_context)

        self.logger = Logging.get_logger('hape.services.kubernetes_model')
        self.core_v1 = kubernetes.client.CoreV1Api()
        self.apps_v1 = kubernetes.client.AppsV1Api()
        self.autoscaling_v2 = kubernetes.client.AutoscalingV2Api()

    def get_deployments(self, namespace):
        self.logger.debug(f"get_deployments(namespace: {namespace})")
        try:
            deployments = self.apps_v1.list_namespaced_deployment(namespace)
            return [deployment.metadata.name for deployment in deployments.items]
        except kubernetes.client.rest.ApiException as e:
            print(f"Error listing deployments in namespace {namespace}: {e}")
            exit(1)

    def get_deployment_replicas(self, namespace):
        self.logger.debug(f"get_deployment_replicas(namespace: {namespace})")
        try:
            deployments = self.apps_v1.list_namespaced_deployment(namespace)
            return {deployment.metadata.name: deployment.spec.replicas for deployment in deployments.items}
        except kubernetes.client.rest.ApiException as e:
            print(f"Error fetching deployment replicas in namespace {namespace}: {e}")
            exit(1)
    
    def get_deployment_cost_details(self, namespace):
        self.logger.debug(f"get_deployment_cost_details(namespace: {namespace})")
        try:
            deployments = self.apps_v1.list_namespaced_deployment(namespace)
            cost_details = {}

            for deployment in deployments.items:
                name = deployment.metadata.name
                replicas = deployment.spec.replicas or 0
                cpu_limit = "N/A"
                ram_limit = "N/A"

                if deployment.spec.template.spec.containers:
                    for container in deployment.spec.template.spec.containers:
                        if container.resources and container.resources.limits:
                            cpu_limit = container.resources.limits.get("cpu", "N/A")
                            ram_limit = container.resources.limits.get("memory", "N/A")
                            break  # Take limits from the first container

                cost_details[name] = {
                    "replicas": replicas,
                    "cpu": cpu_limit,
                    "ram": ram_limit
                }
            return cost_details
        
        except kubernetes.client.rest.ApiException as e:
            print(f"Error fetching deployment cost details in namespace {namespace}: {e}")
            exit(1)

    def remove_hpa_downscaling_annotations(self, hpa_item):
        self.logger.debug(f"remove_hpa_downscaling_annotations(hpa_item: {hpa_item})")
        if not hpa_item.metadata.annotations:
            return
        
        annotations = hpa_item.metadata.annotations.copy()
        keys_to_remove = ["downscaler/downtime-replicas", "downscaler/uptime"]

        for key in keys_to_remove:
            annotations.pop(key, None)

        metadata = kubernetes.V1ObjectMeta(annotations=annotations)
        body = kubernetes.V2HorizontalPodAutoscaler(metadata=metadata)

        try:
            self.autoscaling_v2.patch_namespaced_horizontal_pod_autoscaler(
                name=hpa_item.metadata.name,
                namespace=hpa_item.metadata.namespace,
                body=body
            )
        except kubernetes.client.rest.ApiException as e:
            print(f"Error removing HPA annotations in {hpa_item.metadata.namespace}/{hpa_item.metadata.name}: {e}")

    def remove_hpa_downscaling_annotations_namespaced(self, namespace):
        self.logger.debug(f"remove_hpa_downscaling_annotations_namespaced(namespace: {namespace})")
        try:
            result = self.autoscaling_v2.list_namespaced_horizontal_pod_autoscaler(namespace)
            for item in result.items:
                self.remove_hpa_downscaling_annotations(item)
        except kubernetes.client.rest.ApiException as e:
            print(f"Error listing HPAs in namespace {namespace}: {e}")

    def add_hpa_downscaling_annotations(self, hpa_item):
        self.logger.debug(f"add_hpa_downscaling_annotations(hpa_item: {hpa_item})")
        annotations = hpa_item.metadata.annotations or {}
        annotations["downscaler/downtime-replicas"] = "1"
        annotations["downscaler/uptime"] = "Mon-Fri 05:30-20:30 Europe/Berlin"

        metadata = kubernetes.V1ObjectMeta(annotations=annotations)
        body = kubernetes.V2HorizontalPodAutoscaler(metadata=metadata)

        try:
            self.autoscaling_v2.patch_namespaced_horizontal_pod_autoscaler(
                name=hpa_item.metadata.name,
                namespace=hpa_item.metadata.namespace,
                body=body
            )
        except kubernetes.client.rest.ApiException as e:
            print(f"Error adding HPA annotations in {hpa_item.metadata.namespace}/{hpa_item.metadata.name}: {e}")

    def add_hpa_downscaling_annotations_namespaced(self, namespace):
        self.logger.debug(f"add_hpa_downscaling_annotations_namespaced(namespace: {namespace})")
        try:
            result = self.autoscaling_v2.list_namespaced_horizontal_pod_autoscaler(namespace)
            for item in result.items:
                self.add_hpa_downscaling_annotations(item)
        except kubernetes.client.rest.ApiException as e:
            print(f"Error listing HPAs in namespace {namespace}: {e}")



from hape.logging import Logging

from sqlalchemy import Column, Integer, String, Float, Boolean, BigInteger
from datetime import datetime
from hape.base.model import Model

class DeploymentCost(Model):
    __tablename__ = 'deployment_costs'

    id = Column(Integer, primary_key=True, autoincrement=True)
    created_at = Column(BigInteger, default=lambda: int(datetime.utcnow().timestamp()), nullable=False)
    service_name = Column(String(255), nullable=False)
    pod_cpu = Column(String(50), nullable=False)
    pod_ram = Column(String(50), nullable=False)
    autoscaling = Column(Boolean, nullable=False)
    min_replicas = Column(Integer, nullable=True)
    max_replicas = Column(Integer, nullable=True)
    current_replicas = Column(Integer, nullable=False)
    pod_cost = Column(Float, nullable=False)
    number_of_pods = Column(Integer, nullable=False)
    total_cost = Column(Float, nullable=False)
    cost_unit = Column(String(50), nullable=False)

    def __init__(self, **kwargs):
        filtered_kwargs = {key: kwargs[key] for key in self.__table__.columns.keys() if key in kwargs}
        super().__init__(**filtered_kwargs)
        for key, value in filtered_kwargs.items():
            setattr(self, key, value)



import json
from datetime import datetime
from hape.models.deployment_cost_model import DeploymentCost
from hape.controllers.deployment_cost_controller import DeploymentCostController
from hape.services.gitlab_service import GitlabService
from hape.services.file_service import FileService

class Playground:

    deployment_cost_controller = DeploymentCostController()

    @classmethod
    def main(self):
        playground = Playground()
        playground.play()
        
    def get_all_deployment_costs(self):
        deployment_costs = DeploymentCost.get_all()
        print(DeploymentCost.list_to_json(deployment_costs))

    def save_deployment_cost(self):
        deployment_cost = DeploymentCost(
            service_name="Test Service",
            pod_cpu="2",
            pod_ram="4Gi",
            autoscaling=True,
            min_replicas=1,
            max_replicas=5,
            current_replicas=3,
            pod_cost=0.10,
            number_of_pods=3,
            total_cost=0.30,
            cost_unit="$"
        )
        if self.deployment_cost_controller.save(deployment_cost):
            print('-----saved successfully-----')
        else:
            print('-----saving failed-----')
        
        print(deployment_cost.json())

    def delete_deployment_cost(self):
        deployment_cost = DeploymentCost.get(id=2)
        if not deployment_cost:
            print("Object id=2 does not exist!")
            return
        print(deployment_cost.json())
        deployment_cost.delete()

    def delete_all_deployment_cost(self):
        print("delete all where id in [1,4] and service_name='Test Service'")
        DeploymentCost.delete_all(id=["1", "4"], service_name="Test Service")

    def generate_gitlab_changes_report(self):
        gitlab = GitlaService()
        start_date = datetime(2025, 2, 3)
        end_date = datetime(2025, 2, 5)
        gitlab.generate_csv_changes_in_cicd_repos(
            group_id=178,
            start_date=start_date,
            end_date=end_date,
            output_file="/Users/hazemataya/Desktop/workspace/innodp/playground/test.csv",
            file_regex=r".*values.*.yaml"
        )

    def play(self):
        Playground().save_deployment_cost()
        Playground().get_all_deployment_costs()
        Playground().delete_deployment_cost()
        Playground().delete_all_deployment_cost()
        Playground().generate_gitlab_changes_report()



from abc import ABC, abstractmethod
from sqlalchemy import Integer, String, Float, Boolean

class ModelArgumentParser(ABC):

    _sqlalchemy_type_map = {
        Integer: int,
        String: str,
        Float: float,
        Boolean: lambda x: x.lower() in ('true', '1', 'yes')
    }

    def __init__(self, base_model_class, controller_class):
        self._base_model_class = base_model_class
        self._base_model_name = base_model_class.__name__
        self._base_model_command = ''.join(['-' + char.lower() if char.isupper() else char for char in base_model_class.__name__]).lstrip('-')
        self._base_model_columns = {
            column.name: self._sqlalchemy_type_map.get(type(column.type), str)
            for column in self._base_model_class.__table__.columns
        }
        self._controller = controller_class()
        self.base_model_subparser = None
        self.args = None
    
    @abstractmethod
    def extend_subparser(self):
        pass
    
    @abstractmethod
    def extend_actions(self):
        pass
    
    def create_subparser(self, subparsers):  
        base_model_parser = subparsers.add_parser(self._base_model_command, help=f"Commands to manage {self._base_model_name} base_model")
        self.base_model_subparser = base_model_parser.add_subparsers(dest="action")

        for action in ["save", "get", "get-all", "delete", "delete-all"]:
            object_word = "objects" if "-all" in action else "object"
            parser = self.base_model_subparser.add_parser(action, help=f"{action.capitalize()} {self._base_model_name} {object_word} based on passed arguments or filters")
            
            for column_name, column_type in self._base_model_columns.items():
                column_name_dashes = column_name.replace('_', '-')
                parser.add_argument(f"--{column_name_dashes}", required=False, help=f"Value for {column_name_dashes} type {column_type.__name__}")

        self.extend_subparser()
    
    def run_action(self, args):
        
        self.args = args
        if args.command != self._base_model_command:
            return
        
        filters = {}
        for column_name, column_type in self._base_model_columns.items():
            if hasattr(args, column_name) and getattr(args, column_name):
                try:
                    filters[column_name] = column_type(getattr(args, column_name))
                except ValueError as e:
                    raise ValueError(f"Error casting {column_name}: {str(e)}")
        
        if args.action == "save":
            base_model = self._base_model_class(**filters)
            if not base_model.validate():
                raise ValueError("Model validation failed.")
            self._controller.save(base_model)
            print(base_model.json())
        
        elif args.action == "get":
            base_model = self._controller.get(**filters)
            if not base_model:
                print(f"{self._base_model_name} object not found.")
                return
            print(base_model.json())
        
        elif args.action == "get-all":
            base_model_list = self._controller.get_all(**filters)
            if not base_model_list:
                print(f"No {self._base_model_name} objects found.")
                return
            print(self._base_model_class.list_to_json(base_model_list))
        
        elif args.action == "delete":
            base_model = self._controller.get(**filters)
            if not base_model:
                print(f"{self._base_model_name} object not found.")
                return
            self._controller.delete(base_model)
            print("Deleted object:")
            print(base_model.json())
        
        elif args.action == "delete-all":
            base_model_list = self._controller.get_all(**filters)
            if not base_model_list:
                print(f"No {self._base_model_name} objects found.")
                return
            for base_model in base_model_list:
                self._controller.delete(base_model)
            print("Deleted objects:")
            print(self._base_model_class.list_to_json(base_model_list))
        
        else:
            self.extend_actions()



from hape.logging import Logging
from hape.controllers.git_controller import GitController

class GitArgumentParser:

    def __init__(self):
        logger = Logging.get_logger('hape.argument_parsers.git_argument_parser')

    def create_subparser(self, subparsers):    
        self.logger.debug(f"create_subparser(subparsers: {subparsers})")
        git_parser = subparsers.add_parser("git", help="Commands related to git")
        git_parser_subparser = git_parser.add_subparsers(dest="action")

        clone_parser = git_parser_subparser.add_parser("clone", help="Clone all projects in the specified group")
        clone_parser.add_argument("-d", "--dir", required=True, help="Directory where projects will be cloned")
        clone_parser.add_argument("-g", "--group-id", required=True, type=int, help="GitLab group ID.")

        commit_parser = git_parser_subparser.add_parser("commit", help="Commit the git repositories in the specified directory")
        commit_parser.add_argument("-d", "--dir", required=True, help="Directory where projects will be committed")
        commit_parser.add_argument("-m", "--message", required=True, help="Commit message")
        commit_parser.add_argument("-p", "--prefix", required=False, help="A prefix for the project names", default='cicd-')

        pull_parser = git_parser_subparser.add_parser("pull", help="Pull and update the git repositories in the specified directory using the already checked out branch")
        pull_parser.add_argument("-d", "--dir", required=True, help="Directory where projects will be pulled")
        pull_parser.add_argument("-p", "--prefix", required=False, help="A prefix for the project names", default='')

    def run_action(self, args):
        self.logger.debug(f"run_action(args: {args})")
        if args.command != "git":
            return
        
        controller = GitController()

        if args.action == "clone":
            controller.clone_project_in_group(args.dir, args.group_id)
        elif args.action == "commit":
            controller.commit_projects(args.dir, args.message, args.prefix)
        elif args.action == "pull":
            controller.pull_projects(args.dir, args.prefix)
        else:
            print(f"Error: Invalid {args.command} action. Use `hape {args.command} --help` for more details.")
            exit(1)



######################
######################
# You must not include config.py to avoid circular import.
# Config class uses loggers
######################
import os
import sys
import shutil
import logging
from datetime import datetime
from pythonjsonlogger import jsonlogger

GLOBAL_LOGGER_NAME = "hape.global"

class CustomJsonFormatter(jsonlogger.JsonFormatter):
    def parse(self):
        return ["timestamp", "level", "name", "message", "module", "funcName", "lineno"]

LOGGING_CONFIG = '{"version":1,"disable_existing_loggers":false,"formatters":{"standard":{"format":"%(asctime)s %(levelname)s - %(name)s - %(message)s"}},"handlers":{"console":{"class":"logging.StreamHandler","formatter":"standard","level":"{{log_level}}"},"file":{"class":"logging.handlers.RotatingFileHandler","formatter":"json","level":"{{log_level}}","filename":"{{log_file}}","maxBytes":10485760,"backupCount":5}},"loggers":{"":{"handlers":["console","file"],"level":"{{log_level}}","propagate":false},"uvicorn":{"handlers":["console","file"],"level":"{{log_level}}","propagate":false}}}'

class Logging:

    @staticmethod
    def get_logger(name=GLOBAL_LOGGER_NAME):
        logger = logging.getLogger(name)
        if "--version" in sys.argv:
            logging.disable(logging.CRITICAL)
        return logger

    @staticmethod
    def rotate_log_file(log_file):
        if os.path.exists(log_file):
            timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
            file_name, file_ext = os.path.splitext(log_file)
            file_ext = ".log"
            new_log_file = f"{file_name}_{timestamp}{file_ext}"
            shutil.move(log_file, new_log_file)



import sys
from hape.playground import Playground

class PlaygroundArgumentParser:

    def create_subparser(self, subparsers):    
        if len(sys.argv) > 1 and sys.argv[1] == "play":
            play_parser = subparsers.add_parser("play")
            play_parser.description = ""        

    def run_action(self, args):
        if args.command != "play":
            return
        
        Playground.main()
    


from hape.logging import Logging
from hape.hape_cli.controllers.init_controller import InitController

class InitArgumentParser:
    def __init__(self):
        self.COMMAND = "init"
        self.logger = Logging.get_logger('hape.hape_cli.argument_parsers.init_argument_parser')

    def create_subparser(self, subparsers):    
        self.logger.debug(f"create_subparser(subparsers: {subparsers})")
        init_parser = subparsers.add_parser(self.COMMAND, help="Commands related to initializing the workspace")
        init_parser_subparser = init_parser.add_subparsers(dest="action")

        project_parser = init_parser_subparser.add_parser("project", help="Initializes a new project")
        project_parser.add_argument("-n", "--name", required=True, help="Name of the project")

    def run_action(self, args):
        self.logger.debug(f"run_action(args: {args})")
        if args.command != self.COMMAND:
            return
        if args.action == "project":
            InitController(args.name).init_project()
        else:
            print(f"Error: Invalid {args.command} action. Use `hape {args.command} --help` for more details.")
            exit(1)



import os
import re
import sys
import importlib.util
from hape.logging import Logging
from hape.services.file_service import FileService
from hape.hape_cli.init_structure.new_project_structure import NEW_PROJECT_STRUCTURE

class Init:

    def __init__(self, name: str):
        self.logger = Logging.get_logger('hape.hape_cli.models.init_model')
        self.name = name
        self.name_underscore = name.replace("-", "_")
        self.file_service = FileService()
        
        spec = importlib.util.find_spec("hape")
        if spec and spec.origin:
            self.hape_framework_path = os.path.dirname(os.path.abspath(spec.origin))
        else:
            self.logger.error("Couldn't not find `hape` package. Execute `pip install --upgrade hape`.")
            exit(1)

    def validate(self):
        self.logger.debug(f"validate()")
        name = self.name.strip()
        self.logger.debug(f'Validating project name: {self.name}')
        if not re.match(r'^[a-z0-9]+(-[a-z0-9]+)*$', name):
            self.logger.error(f"Error: Project name '{name}' must contain only lowercase letters, numbers, and use '-' as a separator.")
            sys.exit(1)
        self.logger.debug(f'Valid project name.')

    def _init_file(self, path: str, content: str):
        self.logger.debug(f"_init_file(path: {path}, content: {content})")
        if content:
            content = content.replace("{{project_name}}", self.name_underscore)
        self.file_service.write_file(path, content)

    def _init_directory(self, root_path: str, dictionary: dict):
        self.logger.debug(f"_init_directory(root_path: {root_path}, dictionary: {dictionary})")
        for name, content in dictionary.items():
            if name == 'project_name':
                name = self.name_underscore
            sub_path = os.path.join(root_path, name)
            if content is None:
                self._init_file(sub_path,"")
            elif isinstance(content, str):
                self._init_file(sub_path, content)
            elif isinstance(content, dict):
                self.file_service.create_directory(sub_path)
                self._init_directory(sub_path, content)
            else:
                self.logger.error(f"Content type for {content} is not supported.")
                exit(1)

    def _init_project_structure(self):
        self.logger.debug(f"_init_project_structure()")
        if self.file_service.path_exists(self.name):
            self.logger.warning(f"Warning: directory '{self.name}' already exists.")
            user_input = input(f"ALL DATA IN '{self.name}' WILL BE LOST. Do you want to remove '{self.name}' and initialize a new project? yes/no: ").strip().lower()
            if user_input == 'yes':
                self.file_service.delete_directory(self.name)
            else:
                print("Operation canceled. Keeping the directory.")
                sys.exit(1)
        self.file_service.create_directory(self.name)
        self._init_directory(self.name, NEW_PROJECT_STRUCTURE)

    def init_project(self):
        self.logger.debug(f"init_project()")
        self._init_project_structure()
        self.logger.info(f'Project {self.name} has been successfully initialized!')
        print(f'Project {self.name} has been successfully initialized!')




from hape.hape_cli.models.init_model import Init

class InitController:

    def __init__(self, name):
        self.__init = Init(name)
        self.__init.validate()
    
    def init_project(self):
        self.__init.init_project()



from hape.bootstrap import bootstrap
from hape.hape_cli.argument_parsers.main_argument_parser import MainArgumentParser

def main():
    bootstrap()
    
    main_parser = MainArgumentParser()
    
    parser = main_parser.create_parser()
    args = parser.parse_args()

    if args.command is None:
        parser.print_help()
        return

    main_parser.run_action(args)



from hape.base.model_argument_parser import ModelArgumentParser
from hape.models.deployment_cost_model import DeploymentCost
from hape.controllers.deployment_cost_controller import DeploymentCostController

class DeploymentCostArgumentParser(ModelArgumentParser):
    def __init__(self):
        super().__init__(DeploymentCost, DeploymentCostController)

    def extend_subparser(self):
        pass
    
    def extend_actions(self):
        pass



BOOTSTRAP_PY = """
def bootstrap():
    pass
""".strip()


MAIN_PY = """
from {{project_name}}.cli import main

if __name__ == "__main__":
    main()
""".strip()


PLAYGROUND_PY = """
class Playground:

    @classmethod
    def main(self):
        playground = Playground()
        playground.play()

    def play(self):
        print("Playground.play() ran successfully!")
""".strip()


CLI_PY = """
from {{project_name}}.bootstrap import bootstrap
from {{project_name}}.argument_parsers.main_argument_parser import MainArgumentParser

def main():
    bootstrap()
    
    main_parser = MainArgumentParser()
    
    parser = main_parser.create_parser()
    args = parser.parse_args()

    if args.command is None:
        parser.print_help()
        return

    main_parser.run_action(args)
""".strip()


MAIN_ARGUMENT_PARSER = """
import argparse
from importlib.metadata import version

from {{project_name}}.argument_parsers.playground_argument_parser import PlaygroundArgumentParser

class MainArgumentParser:

    def create_parser(self):
        parser = argparse.ArgumentParser(
            description="{{project_name}} created by HAPE Framework"
        )
        try:
            parser.add_argument("-v", "--version", action="version", version=version("{{project_name}}"))
        except:
            parser.add_argument("-v", "--version", action="version", version="0.0.0")
        
        subparsers = parser.add_subparsers(dest="command")
        
        PlaygroundArgumentParser().create_subparser(subparsers)
        return parser
    
    def run_action(self, args):
        if args.command == "play":
            PlaygroundArgumentParser().run_action(args)
        else:
            print(f"Error: Invalid command {args.command}")
            exit(1)
""".strip()

PLAYGROUND_ARGUMENT_PARSER = """
import sys
from {{project_name}}.playground import Playground

class PlaygroundArgumentParser:
    def create_subparser(self, subparsers):    
        play_parser = subparsers.add_parser("play")
        play_parser.description = "Runs Playground.play() function"        

    def run_action(self, args):
        if args.command != "play":
            return
        
        Playground.main()
"""

MIGRATION_ENV_PY = """
from logging.config import fileConfig

from sqlalchemy import engine_from_config
from sqlalchemy import pool

from alembic import context

# this is the Alembic Config object, which provides
# access to the values within the .ini file in use.
config = context.config

# Interpret the config file for Python logging.
# This line sets up loggers basically.
if config.config_file_name is not None:
    fileConfig(config.config_file_name)

# add your model's MetaData object here
# for 'autogenerate' support
# from myapp import mymodel
# target_metadata = mymodel.Base.metadata
from hape.base.model import Base  # Modify this to match your project structure
target_metadata = Base.metadata

# other values from the config, defined by the needs of env.py,
# can be acquired:
# my_important_option = config.get_main_option("my_important_option")
# ... etc.


def run_migrations_offline() -> None:
    url = config.get_main_option("sqlalchemy.url")
    context.configure(
        url=url,
        target_metadata=target_metadata,
        literal_binds=True,
        dialect_opts={"paramstyle": "named"},
    )

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online() -> None:
    connectable = engine_from_config(
        config.get_section(config.config_ini_section, {}),
        prefix="sqlalchemy.",
        poolclass=pool.NullPool,
    )

    with connectable.connect() as connection:
        context.configure(
            connection=connection, target_metadata=target_metadata
        )

        with context.begin_transaction():
            context.run_migrations()


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()
"""


from hape.logging import Logging

import subprocess
from dotenv import load_dotenv
import os
import requests
from hape.hape_config import HapeConfig

class GitService:
    
    def __init__(self):
        self.logger = Logging.get_logger('hape.services.git_service')
        self.dry_run = False
        self.gitlab_token = HapeConfig.get_gitlab_token()
        gitlab_domain = HapeConfig.get_gitlab_domain()
        self.gitlab_url = f"https://{gitlab_domain}"

    def git_branch_name(self, repo_dir):
        self.logger.debug(f"git_branch_name(repo_dir: {repo_dir})")
        result = subprocess.run(['git', 'rev-parse', '--abbrev-ref', 'HEAD'], cwd=repo_dir, capture_output=True, text=True)
        if result.returncode != 0:
            raise Exception(f"Failed to determine the current branch in {repo_dir}: {result.stderr}")
        return result.stdout.strip()

    def git_pull(self, repo_dir, branch_name):
        self.logger.debug(f"git_pull(repo_dir: {repo_dir}, branch_name: {branch_name})")
        print(f"$ git pull origin {branch_name}")
        if not self.dry_run:
            subprocess.run(['git', 'pull', 'origin', branch_name], cwd=repo_dir, check=True)

    def git_has_changes(self, repo_dir):
        self.logger.debug(f"git_has_changes(repo_dir: {repo_dir})")
        result = subprocess.run(['git', 'status', '--porcelain'], cwd=repo_dir, capture_output=True, text=True)
        if result.stdout:
            return True
        else:
            return False

    def git_add(self, repo_dir):
        self.logger.debug(f"git_add(repo_dir: {repo_dir})")
        print(f"$ git add .")
        if not self.dry_run:
            subprocess.run(['git', 'add', '.'], cwd=repo_dir, check=True)
        
    def git_review_and_commit(self, repo_dir):
        self.logger.debug(f"git_review_and_commit(repo_dir: {repo_dir})")
        print("review and commit")
        if not self.dry_run:
            subprocess.run(['git-cola'], cwd=repo_dir, check=True)
    
    def git_commit(self, repo_dir, commit_message):
        self.logger.debug(f"git_commit(repo_dir: {repo_dir}, commit_message: {commit_message})")
        print(f"commit changes, message '{commit_message}'")
        if not self.dry_run:
            subprocess.run(['git', 'commit', '-am', commit_message], cwd=repo_dir, check=True)

    def git_push(self, repo_dir, branch_name = None):
        self.logger.debug(f"git_push(repo_dir: {repo_dir}, branch_name : {branch_name })")
        if not branch_name:
            branch_name = self.git_branch_name(repo_dir)
        print(f"$ git push origin {branch_name}")
        if not self.dry_run:
            subprocess.run(['git', 'push', 'origin', branch_name], cwd=repo_dir, check=True)
    
    def git_clone(self, clone_url, project_path):
        self.logger.debug(f"git_clone(clone_url: {clone_url}, project_path: {project_path})")
        print(f"$ git clone {clone_url} {project_path}")
        subprocess.run(['git', 'clone', clone_url, project_path], check=True)
        
    def get_projects_in_group(self, group_id):
        self.logger.debug(f"get_projects_in_group(group_id: {group_id})")
        headers = {'Private-Token': self.gitlab_token}
        page = 1
        all_projects = []
        while True:
            projects_url = f"{self.gitlab_url}/api/v4/groups/{group_id}/projects"
            params = {'include_subgroups': 'true', 'archived': 'false', 'simple': 'true', 'per_page': '100', 'page': page}
            response = requests.get(projects_url, headers=headers, params=params)
            if response.status_code != 200:
                raise Exception(f"Failed to fetch projects: {response.json().get('message', 'No error message')}")
            projects = response.json()
            if not projects:
                break  # Break the loop if no more projects are returned
            
            all_projects = all_projects + projects
            page += 1

        return all_projects
    
    def get_all_projects(self):
        self.logger.debug(f"get_all_projects(self: {self})")
        headers = {'Private-Token': self.gitlab_token}
        page = 1
        all_projects = []
        while True:
            projects_url = f"{self.gitlab_url}/api/v4/projects"
            params = {'include_subgroups': 'true', 'archived': 'false', 'simple': 'true', 'per_page': '100', 'page': page}
            response = requests.get(projects_url, headers=headers, params=params)
            if response.status_code != 200:
                raise Exception(f"Failed to fetch projects: {response.json().get('message', 'No error message')}")
            projects = response.json()
            if not projects:
                break  # Break the loop if no more projects are returned
            
            all_projects = all_projects + projects
            page += 1

        return all_projects
            
    def clone_all_projects(self, group_id, clone_dir):
        self.logger.debug(f"clone_all_projects(group_id: {group_id}, clone_dir: {clone_dir})")
        projects = self.get_projects_in_group(group_id)
        for project in projects:
            clone_url = project['ssh_url_to_repo']
            project_path = os.path.join(clone_dir, project['path_with_namespace'])
            if not os.path.exists(project_path):
                print(f"Cloning {project['name']}...")
                self.git_clone(clone_url, project_path)
            else:
                print(f"{project_path} already exists in {clone_dir}. Skipping clone.")

    def get_git_repositories(self, dir, prefix, recursive=True):
        self.logger.debug(f"get_git_repositories(dir: {dir}, prefix: {prefix}, recursive: {recursive})")
        subdirectories = []
        if recursive:
            for root, dirs, _ in os.walk(dir):
                for subdir in sorted(dirs):
                    full_path = os.path.join(root, subdir)
                    if os.path.isdir(full_path) and \
                       os.path.exists(os.path.join(full_path, ".git")):
                        if prefix and subdir.startswith(prefix):
                            subdirectories.append(full_path)
                        else:
                            subdirectories.append(full_path)
        else:
            subdirectories = [
                os.path.join(dir, subdir) for subdir in sorted(os.listdir(dir))
                if subdir.startswith(prefix) and \
                    os.path.isdir(os.path.join(dir, subdir)) and \
                    os.path.exists(os.path.join(dir, subdir, ".git"))
                ]
        return subdirectories


SETUP_PY = """
from setuptools import setup, find_packages

setup(
    name="{{project_name}}",
    version="0.0.1",
    packages=find_packages(include=["{{project_name}}", "{{project_name}}/*"]),
    package_data={},
    include_package_data=True,
    install_requires=[
        
    ],
    entry_points={
        "console_scripts": [
            "{{project_name}}={{project_name}}.{{project_name}}_cli.cli:main",
        ],
    },
    author="Hazem Ataya",
    author_email="hazem.ataya94@gmail.com",
    description="HAPE Framework: Build an Automation Tool With Ease",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/hazemataya94/hape-framework",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.9",
)

""".strip()


from setuptools import setup, find_packages

setup(
    name="hape",
    version="0.2.64",
    packages=["hape"],
    include_package_data=True,
    install_requires=[
        "python-dotenv==1.0.1"
    ],
    entry_points={
        "console_scripts": [
            "hape=hape.hape_cli.cli:main",
        ],
    },
    author="Hazem Ataya",
    author_email="hazem.ataya94@gmail.com",
    description="HAPE Framework: Build an Automation Tool With Ease",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/hazemataya94/hape-framework",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.9",
)



import json
from hape.models.git_model import GitModel
from hape.logging import Logging

class GitController:
    def __init__(self):
        self.logger = Logging.get_logger('hape.controllers.git_controller')
        self.git_model = GitModel()
    
    def clone_project_in_group(self, dir, group_id):
        self.logger.debug(f"clone_project_in_group(dir: {dir}, group_id: {group_id})")
        return self.git_model.clone_project_in_group(dir, group_id)
    
    def commit_projects(self, dir, message, prefix):
        self.logger.debug(f"commit_projects(dir: {dir}, message: {message}, prefix: {prefix})")
        repos = self.git_model.get_repositories(dir, prefix)
        
        print("The following projects will be committed and pushed:")
        print(json.dumps(repos, indent=2))
        print(f"Total number of affected projects: {len(repos)}")
        
        user_input = input("Do you want to continue? (y/n): ").strip().lower()
        if user_input != "y":
            print("Aborting task...")
            return
        
        for repo in repos:
            print("==============================")
            print(f"Directory: {repo}")
            if self.git_model.commit_and_push(repo, message):
                print("Changes committed and pushed successfully.")
            else:
                print("No changes to commit.")
    
    def pull_projects(self, dir, prefix):
        self.logger.debug(f"pull_projects(dir: {dir}, prefix: {prefix})")
        repos = self.git_model.get_repositories(dir, prefix)
        
        print("The following projects will be pulled:")
        print(json.dumps(repos, indent=2))
        print(f"Total number of affected projects: {len(repos)}")
        
        for repo in repos:
            print("==============================")
            print(f"Directory: {repo}")
            self.git_model.pull_project(repo)
            print("Repository updated successfully.")



from hape.logging import Logging

from hape.base.model_controller import ModelController
from hape.models.deployment_cost_model import DeploymentCost

class DeploymentCostController(ModelController):
    
    def __init__(self):
        super().__init__(DeploymentCost)



DOCKERFILE_DEV = """
FROM python:3.12-bookworm

WORKDIR /workspace

COPY . /workspace

WORKDIR /workspace/playground

ENTRYPOINT ["sleep", "infinity"]
""".strip()

DOCKERFILE_PROD = """
FROM python:3.12-bookworm

RUN pip install --upgrade {{project_name}}

ENTRYPOINT ["{{project_name}}"]
""".strip()


DOCKER_COMPOSE = """
services:
  {{project_name}}:
    build:
      context: .
      dockerfile: Dockerfile.dev
    container_name: {{project_name}}
    restart: always
    env_file:
      - ../.env
    environment:
      HAPE_MARIADB_HOST: "mariadb"
    networks:
      - host_network

  mariadb:
    image: mariadb:11.4.4
    container_name: mariadb_dev
    restart: always
    environment:
      MARIADB_ROOT_PASSWORD: root
      MARIADB_DATABASE: {{project_name}}_db
      MARIADB_USER: {{project_name}}_user
      MARIADB_PASSWORD: {{project_name}}_password
    ports:
      - "3306:3306"
    networks:
      - host_network
    volumes:
      - mariadb_data:/var/lib/mysql
      - ./mariadb-init:/docker-entrypoint-initdb.d

  phpmyadmin:
    image: phpmyadmin/phpmyadmin
    container_name: phpmyadmin_dev
    restart: always
    environment:
      PMA_HOST: host.docker.internal
      PMA_PORT: 3306
    ports:
      - "8080:80"
    networks:
      - host_network

networks:
  host_network:
    driver: bridge

volumes:
  mariadb_data:
""".strip()


from hape.logging import Logging
from hape.services.git_service import GitService

class GitModel:
    def __init__(self):
        self.logger = Logging.get_logger('hape.models.git_model')
        self._git_service = GitService()
    
    def clone_project_in_group(self, dir, group_id):
        self.logger.debug(f"clone_project_in_group(dir: {dir}, group_id: {group_id})")
        self._git_service.clone_all_projects(group_id, dir)
    
    def get_repositories(self, dir, prefix):
        self.logger.debug(f"get_repositories(dir: {dir}, prefix: {prefix})")
        return self._git_service.get_git_repositories(dir, prefix)
    
    def git_has_changes(self, repo):
        self.logger.debug(f"git_has_changes(repo: {repo})")
        return self._git_service.git_has_changes(repo)
    
    def git_branch_name(self, repo):
        self.logger.debug(f"git_branch_name(repo: {repo})")
        return self._git_service.git_branch_name(repo)
    
    def commit_and_push(self, repo, message):
        self.logger.debug(f"commit_and_push(repo: {repo}, message: {message})")
        if self.git_has_changes(repo):
            branch_name = self.git_branch_name(repo)
            self._git_service.git_pull(repo, branch_name)
            self._git_service.git_add(repo)
            self._git_service.git_commit(repo, message)
            self._git_service.git_push(repo, branch_name)
            return True
        return False
    
    def pull_project(self, repo):
        self.logger.debug(f"pull_project(repo: {repo})")
        branch_name = self.git_branch_name(repo)
        self._git_service.git_pull(repo, branch_name)